import logging

from .base_command import BaseCommand
from ..exception import CTERAException
from ..common import Object, DateTimeUtils
from . import query
from . import union


class Users(BaseCommand):
    """
    Portal User Management APIs
    """

    default = ['name']

    def get(self, user_account, include=None):
        """
        Get a user account

        :param cterasdk.core.types.UserAccount user_account: User account, including the user directory and user name
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The user account, including the requested fields
        """
        baseurl = '/users/%s' % user_account.name if user_account.is_local \
            else '/domains/%s/adUsers/%s' % (user_account.directory, user_account.name)
        include = union.union(include or [], Users.default)
        include = ['/' + attr for attr in include]
        user_object = self._portal.get_multi(baseurl, include)
        if user_object.name is None:
            raise CTERAException('Could not find user', None, user_directory=user_account.directory, username=user_account.name)
        return user_object

    def list_local_users(self, include=None):
        """
        List all local users

        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for all local users
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union.union(include or [], Users.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._portal, '/users', param)

    def list_domains(self):
        """
        List all domains

        :return list: List of all domains
        """
        return self._portal.get('/domains')

    def list_domain_users(self, domain, include=None):
        """
        List all the users in the domain

        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: Iterator for all the domain users
        :rtype: cterasdk.lib.iterator.Iterator
        """
        include = union.union(include or [], Users.default)
        param = query.QueryParamBuilder().include(include).build()
        return query.iterator(self._portal, '/domains/' + domain + '/adUsers', param)

    def add(self, name, email, first_name, last_name, password, role, company=None, comment=None, password_change=False):
        """
        Add a portal user

        :param str name: User name for the new user
        :param str email: E-mail address of the new user
        :param str first_name: The first name of the new user
        :param str last_name: The last name of the new user
        :param str password: Password for the new user
        :param cterasdk.core.enum.Role role: User role of the new user
        :param str,optional company: The name of the company of the new user, defaults to None
        :param str,optional comment: Additional comment for the new user, defaults to None
        :param variable,optional password_change:
            Require the user to change the password on the first login.
            Pass datetime.date for a specific date, integer for days from creation, or True for immediate , defaults to False
        """
        param = Object()
        param._classname = "PortalUser"  # pylint: disable=protected-access
        param.name = name
        param.email = email
        param.firstName = first_name
        param.lastName = last_name
        param.password = password
        param.role = role
        param.company = company
        param.comment = comment
        if password_change:
            param.requirePasswordChangeOn = DateTimeUtils.get_expiration_date(password_change).strftime('%Y-%m-%d')

        logging.getLogger().info('Creating user. %s', {'user': name})
        response = self._portal.add('/users', param)
        logging.getLogger().info('User created. %s', {'user': name, 'email': email, 'role': role})

        return response

    def delete(self, user):
        """
        Delete a user

        :param cterasdk.core.types.UserAccount user: the user account
        """
        logging.getLogger().info('Deleting user. %s', {'user': str(user)})
        baseurl = '/users/%s' % user.name if user.is_local else '/domains/%s/adUsers/%s' % (user.directory, user.name)
        response = self._portal.execute(baseurl, 'delete', True)
        logging.getLogger().info('User deleted. %s', {'user': str(user)})

        return response



import re
import os
import pathlib
from xml.dom import minidom
from xml.sax.saxutils import escape
from xml.parsers.expat import ExpatError


def transcribe(request, response=None):  # pylint: disable=too-many-locals
    comment = '<!--TEMPLATE-->'
    filename = './CTERA_' + str(os.getpid()) + '.html'

    file = None
    if not os.path.exists(filename):
        template = os.path.join(pathlib.Path(__file__).parent.absolute(), 'apidoc.template')
        file = open(template, 'r').read()
    else:
        file = open(filename, 'r').read()

    array = file.split(comment)
    header = array[0]
    footer = array[1]

    uri = request.full_url
    idx = re.search("https?://[^/]*", uri).end()
    uri = uri[idx:]
    if uri.endswith('?'):
        uri = uri[:-1]

    method = request.get_method()
    tuples = request.header_items()
    headers = []
    for key, value in tuples:
        if key in ['Content-type', 'Cookie']:
            headers.append((key, value))

    data = ''
    if request.data is not None:
        data = request.data.decode('utf-8')

    file = open(filename, 'w')

    string = '<div style="border: 1px solid #EBECF0; margin-bottom: 10px;">'
    string = string + '<p class="h4">' + method + ' ' + uri + '</p>'

    # Request headers:
    if len(headers) > 0:
        string = string + "<br/><p class=\"h6\">Request headers:</p>"
        for key, value in headers:
            string = string + '<p>' + key + ': ' + value + '</p>'

    # Request body:
    if data:
        string = string + "<br/><p class=\"h6\">Request body:</p>"
        string = string + '<pre>' + prettify(data) + '</pre>'

    # Response body:
    if response is not None:
        reply = response.read()
        if reply:
            string = string + "<br/><p class=\"h6\">Response body:</p>"
            string = string + '<pre>' + prettify(reply) + '</pre>'

    string = string + '</div>'
    file.write(header + string + comment + footer)
    file.close()


def prettify(data):
    try:
        data = minidom.parseString(data).toprettyxml(indent="   ")
        data = ''.join(data.split('\n', 1)[1:])
        data = escape(data)
        return data
    except ExpatError:
        return data



class Context:
    """
    Portal connection context

    :ivar str admin: Global admin context
    :ivar str ServicesPortal: Services Portal context
    """
    admin = 'admin'
    ServicesPortal = 'ServicesPortal'


class LogTopic:
    """
    Portal Log Topic

    :ivar str System: System log topic
    :ivar str CloudBackup: Cloud Backup log topic
    :ivar str CloudSync: Cloud Sync log topic
    :ivar str Access: Access log topic
    :ivar str Audit: Audit log topic
    """
    System = 'system'
    CloudBackup = 'backup'
    CloudSync = 'cloudsync'
    Access = 'access'
    Audit = 'audit'


class OriginType:
    """
    Log Origin Type

    :ivar str Portal: Portal originated logs
    :ivar str Device: Device originated logs
    """
    Portal = 'Portal'
    Device = 'Device'


class DeviceType:
    """
    Device type

    :ivar str CloudPlug: Cloud Plug device
    :ivar str C200: C200 device
    :ivar str C400: C400 device
    :ivar str C800: C800 device
    :ivar str C800P: C800P device
    :ivar str vGateway: vGateway device
    :ivar str ServerAgent: Server Agent device
    :ivar str WorkstationAgent: Workstation Agent Agent device
    :ivar list[str] Gateways: List of all the Gateway DeviceTypes
    :ivar list[str] Agents: List of all the Agents DeviceTypes
    """
    CloudPlug = "CloudPlug"
    C200 = "C200"
    C400 = "C400"
    C800 = "C800"
    C800P = "C800P"
    vGateway = "vGateway"
    ServerAgent = "Server Agent"
    WorkstationAgent = "Workstation Agent"
    Gateways = [CloudPlug, C200, C400, C800, C800P, vGateway]
    Agents = [ServerAgent, WorkstationAgent]


class Role:
    """
    Portal User Role

    :ivar str Disabled: Disabled user role
    :ivar str EndUser: EndUser user role
    :ivar str ReadWriteAdmin: ReadWriteAdmin user role
    :ivar str ReadOnlyAdmin: ReadOnlyAdmin user role
    :ivar str Support: Support user role
    """
    Disabled = "Disabled"
    EndUser = "EndUser"
    ReadWriteAdmin = "ReadWriteAdmin"
    ReadOnlyAdmin = "ReadOnlyAdmin"
    Support = "Support"


class Severity:
    """
    Portal Log Severity

    :ivar str EMERGENCY: Emergency log severity
    :ivar str ALERT: Alert log severity
    :ivar str CRITICAL: Critical log severity
    :ivar str ERROR: Error log severity
    :ivar str WARNING: Warning log severity
    :ivar str NOTICE: Notice log severity
    :ivar str INFO: Info log severity
    :ivar str DEBUG: Debug log severity
    """
    EMERGENCY = "emergency"
    ALERT = "alert"
    CRITICAL = "critical"
    ERROR = "error"
    WARNING = "warning"
    NOTICE = "notice"
    INFO = "info"
    DEBUG = "debug"


class PortalAccountType:
    """
    Portal Account Type

    :ivar str User: User account type
    :ivar str Group: Group account type
    """
    User = "user"
    Group = "group"


class SearchType:
    """
    Search Type

    :ivar str User: User search type
    :ivar str Group: Group search type
    """
    Users = "users"
    Groups = "groups"


class PolicyType:
    """
    Zone Policy Type

    :ivar str ALL: All folders
    :ivar str SELECT: Selected Folders
    :ivar str NONE: No Folders
    """
    ALL = 'allFolders'
    SELECT = 'selectedFolders'
    NONE = 'noFolders'


class ProtectionLevel:
    """
    External Share Protection Level

    :ivar str publicLink: No authentication
    :ivar str email: 2FA via email
    """
    Public = 'publicLink'
    Email = 'email'


class FileAccessMode:
    """
    Share Access Mode

    :ivar str RW: Read Write
    :ivar str RO: Read Only
    :ivar str PO: Preview Only
    :ivar str NA: None
    """
    RW = "ReadWrite"
    RO = "ReadOnly"
    PO = "PreviewOnly"
    NA = "None"


class CollaboratorType:
    """
    Collaborator Type

    :ivar str LU: Local User
    :ivar str DU: Domain User
    :ivar str LG: Local Group
    :ivar str DG: Domain Group
    :ivar str EXT: External
    """
    LU = "localUser"
    LG = "localGroup"
    DU = "adUser"
    DG = "adGroup"
    EXT = "external"


class PortalType:
    """
    Portal Type

    :ivar str Team: Team Portal
    :ivar str Reseller: Reseller Portal
    """
    Team = 'team'
    Reseller = 'reseller'



import logging
import time
import re

from ..exception import CTERAException, InputError


class TaskStatusEnum:
    Running = "running"
    Failed = "failed"
    Completed = "completed"


class TaskError(CTERAException):

    def __init__(self, task):
        super().__init__()
        self.task = task


class Task:

    def __init__(self, CTERAHost, ref, retries=10, seconds=1):
        self.CTERAHost = CTERAHost
        self.path = self.tid(ref)
        self.attempt = 0
        self.retries = retries
        self.seconds = seconds
        self.running = True

    def wait(self):
        task = None
        while self.running:
            logging.getLogger().debug('Obtaining task status. %s', {'path': self.path, 'attempt': (self.attempt + 1)})
            task = self.CTERAHost.get(self.path)
            self.increment()
            self.running = task.status == TaskStatusEnum.Running
        return self.resolve(task)

    @staticmethod
    def resolve(task):
        if task.status != TaskStatusEnum.Completed:
            logging.getLogger().error('Task did not complete successfully. %s', {'id': task.id, 'status': task.status})
            raise TaskError(task)

        logging.getLogger().debug('Task completed successfully. %s', {'id': task.id})
        return task

    def increment(self):
        if self.attempt >= self.retries:
            duration = time.strftime("%H:%M:%S", time.gmtime(self.retries * self.seconds))
            logging.getLogger().error('Could not obtain task status in a timely manner. %s', {'duration': duration})
            raise CTERAException('Timed out. Could not obtain task status in a timely manner', None, duration=duration)
        self.attempt = self.attempt + 1
        logging.getLogger().debug('Sleep. %s', {'seconds': self.seconds})
        time.sleep(self.seconds)

    @staticmethod
    def tid(ref):
        uid = None
        if isinstance(ref, int):
            uid = str(ref)
        elif isinstance(ref, str):
            match = re.search('[1-9][0-9]*', ref)
            if match is not None:
                start, end = match.span()
                uid = ref[start: end]
        if uid is not None:
            return '/proc/bgtasks/' + uid
        logging.getLogger().error('Could not parse task id. %s', {'ref': ref})
        raise InputError('Invalid task id', ref, [64, '/proc/bgtasks/64'])


def running(CTERAHost):
    return CTERAHost.query('/proc/bgtasks', 'status', 'running')


def by_name(CTERAHost, name):
    return CTERAHost.query('/proc/bgtasks', 'name', name)


def wait(CTERAHost, ref):
    task = Task(CTERAHost, ref)
    return task.wait()



from .base_command import BaseCommand
from ..exception import CTERAException
from . import union


class Plans(BaseCommand):
    """
    Global Admin Plan APIs
    """
    default = ['name']

    def get(self, name, include=None):
        """
        Get a subscription plan

        :param str name: Name of the subscription plan
        :param list[str] include: List of fields to retrieve, defaults to ['name']
        :return: The subscription plan, including the requested fields
        """
        include = union.union(include or [], Plans.default)
        include = ['/' + attr for attr in include]
        plan = self._portal.get_multi('/plans/' + name, include)
        if plan.name is None:
            raise CTERAException('Could not find subscription plan', None, name=name)
        return plan



from datetime import datetime

from .base_command import BaseCommand
from ..lib import Iterator, Command
from ..core import enum
from . import query


class Logs(BaseCommand):
    """
    Portal Logs APIs
    """

    def get(self, topic=enum.LogTopic.System, minSeverity=enum.Severity.INFO, originType=enum.OriginType.Portal, before=None, after=None):
        """
        Get logs from the Portal

        :param cterasdk.core.enum.LogTopic,optional topic: Log topic to get, defaults to cterasdk.core.enum.LogTopic.System
        :param cterasdk.core.enum.Severity,optional minSeverity:
         Minimun severity of logs to get, defaults to cterasdk.core.enum.Severity.INFO
        :param cterasdk.core.enum.OriginType,optional originType:
         Origin type of the logs to get, defaults to cterasdk.core.enum.OriginType.Portal
        :param str,optional before: Get logs before this date (in format "%m/%d/%Y %H:%M:%S"), defaults to None
        :param str,optional after: Get logs after this date (in format "%m/%d/%Y %H:%M:%S"), defaults to None
        """
        builder = query.QueryParamBuilder().put('topic', topic).put('minSeverity', minSeverity)

        builder.addFilter(query.FilterBuilder('originType').eq(originType))

        if before is not None:
            builder.addFilter(query.FilterBuilder('time').before(self._strptime(before)))

        if after is not None:
            builder.addFilter(query.FilterBuilder('time').after(self._strptime(after)))

        param = builder.build()
        function = Command(self._query_logs)

        return Iterator(function, param)

    def _query_logs(self, param):
        response = self._portal.execute('', 'queryLogs', param)
        return (response.hasMore, response.logs)

    @staticmethod
    def _strptime(datetime_str):
        try:
            return datetime.strptime(datetime_str, '%m/%d/%Y %H:%M:%S')
        except ValueError as error:
            raise error



import logging

from . import taskmgr as TaskManager
from ..exception import CTERAException
from .base_command import BaseCommand


class Shell(BaseCommand):
    """ Gateway Shell command """

    def run_command(self, shell_command):
        """
        Execute a shell command on the gateway

        :param str shell_command: The shell command to execute
        :return: The command result
        """
        logging.getLogger().info("Executing shell command. %s", {'shell_command': shell_command})

        task = self._gateway.execute("/config/device", "bgshell", shell_command)
        try:
            task = TaskManager.wait(self._gateway, task)
            logging.getLogger().info("Shell command executed. %s", {'shell_command': shell_command})
            return task.result.result
        except TaskManager.TaskError as error:
            raise CTERAException('An error occurred while executing task', error)



import logging

from ..lib import track, ErrorStatus
from .enum import Mode, SyncStatus, Acl
from .base_command import BaseCommand


class Sync(BaseCommand):
    """ Gateway Cloud Sync APIs """

    def get_status(self):
        """ Retrieve the Cloud Sync status """
        return self._gateway.get('/proc/cloudsync/serviceStatus')

    def is_disabled(self):
        """ Check if Cloud Sync is disabled """
        return self._gateway.get('/config/cloudsync/mode') == Mode.Disabled

    def is_enabled(self):
        """ Check if Cloud Sync is enabled """
        return self._gateway.get('/config/cloudsync/mode') == Mode.Enabled

    def suspend(self):
        """ Suspend Cloud Sync """
        logging.getLogger().info("Suspending cloud sync.")
        self._gateway.put('/config/cloudsync/mode', Mode.Disabled)
        logging.getLogger().info("Cloud sync suspended.")

    def unsuspend(self):
        """ Unsuspend Cloud Sync """
        logging.getLogger().info("Unsuspending cloud sync.")
        self._gateway.put('/config/cloudsync/mode', Mode.Enabled)
        try:
            ref = '/proc/cloudsync/serviceStatus/id'
            track(
                self._gateway,
                ref,
                [
                    SyncStatus.Synced,
                    SyncStatus.Syncing,
                    SyncStatus.Scanning
                ],
                [
                    SyncStatus.ConnectingFolders,
                    SyncStatus.InitializingConnection
                ],
                [
                    SyncStatus.DisconnectedPortal,
                    SyncStatus.Unlicensed,
                    SyncStatus.Off
                ],
                [
                    SyncStatus.ServiceUnavailable,
                    SyncStatus.VolumeUnavailable,
                    SyncStatus.ShouldSupportWinNtAcl,
                    SyncStatus.InternalError,
                    SyncStatus.ClocksOutOfSync,
                    SyncStatus.NoFolder
                ]
            )
            logging.getLogger().info('Unsuspended cloud sync.')
        except ErrorStatus as error:
            if error.status == SyncStatus.ShouldSupportWinNtAcl:
                logging.getLogger().warning('Windows ACL enabled folder cannot be synchronized to this device.')
                self._gateway.put('/config/fileservices/share/cloud/access', Acl.WindowsNT)
                logging.getLogger().info('Updated network share access. %s', {'share': 'cloud', 'access': Acl.WindowsNT})
            else:
                logging.getLogger().error("An error occurred while unsuspendeding sync. %s", {'status': error.status})

    def refresh(self):
        """ Refresh Cloud Folders """
        logging.getLogger().info("Refreshing cloud folders.")
        self._gateway.execute("/config/cloudsync/cloudExtender", "refreshPaths", None)
        logging.getLogger().info("Completed refreshing cloud folders.")
